

# # Programmation du jeu Hexapawn

# ## Modélisation du plateau de jeu et des mouvements

# ### Plateau de jeu




import copy
def affichage(plateau):
    hrule = "   "+'-' * 13
    plateau_aff = copy.deepcopy(plateau) + [["A","B","C"]]
    print(hrule)
    for i in range(len(plateau_aff)):
        ligne = plateau_aff[i]
        if i < 3:
            print(str(3-i)+" ",'|', ' | '.join(cell for cell in ligne), '|')
            print(hrule)
        else:
            print("  ",' ', '   '.join(cell for cell in ligne), '')



# exemple = [["N"," ","N"],[" ","N","B"],["B","B"," "]];
# affichage(exemple) # exécutez la cellule pour avoir le rendu


# ### Etats de jeux et fonctions associées




def coord_vers_chaine(couple):
    """convertit un mouvement d'une liste de deux listes vers une chaine de caractères"""
    cases = [['A3', 'B3', 'C3'], ['A2', 'B2', 'C2'], ['A1', 'B1', 'C1']]
    couples = [[[0,0], [0,1], [0,2]], [[1,0], [1,1],[1,2]], [[2,0], [2,1], [2,2]]]
    conversion = cases[couple[0][0]][couple[0][1]] + cases[couple[1][0]][couple[1][1]]
    return conversion

def chaine_vers_coord(chaine):
    """convertit un mouvement sous la forme d'une chaine de caractères vers une liste de deux listes"""
    abs = []
    ord = []
    cases = [['A3', 'B3', 'C3'], ['A2', 'B2', 'C2'], ['A1', 'B1', 'C1']]
    couples = [[[0,0], [0,1], [0,2]], [[1,0], [1,1],[1,2]], [[2,0], [2,1], [2,2]]]
    for i in range(3):
        for j in range(3):
            if cases[i][j] == chaine[:2]:
                abs = couples[i][j]
            if cases[i][j] == chaine[2:]:
                ord = couples[i][j]
    return [abs,ord]


def direction(couleur):
    """direction de mouvement selon la couleur"""
    direction = -1
    if couleur == "N":
        direction = 1
    elif couleur == "B":
        direction = -1
    return direction

def autre_couleur(couleur):
    """renvoie l'autre couleur"""
    if couleur == "N":
        return "B"
    elif couleur == "B":
        return "N"

def rang_initial(couleur):
    """renvoie le rang initial de la couleur"""
    rang = 1
    if couleur == "N":
        rang = 0
    elif couleur == "B":
        rang = 2
    return rang


def sym_mouv(mouvement):
    """renvoie le symétrique d'un mouvement sous la forme d'un couple"""
    mouv = copy.deepcopy(mouvement)
    lg_depart = mouv[0][0]
    col_depart = mouv[0][1]
    lg_arrivee = mouv[1][0]
    col_arrivee = mouv[1][1]
    return [[lg_depart, 2 - col_depart], [lg_arrivee, 2 - col_arrivee]]


def sym_chaine(chaine):
    """renvoie le symétrique d'un mouvement en échangeant A et C"""
    sym = ""
    for lettre in chaine:
        if lettre == 'A':
            car = 'C'
        elif lettre == 'C':
            car = 'A'
        else:
            car = lettre
        sym = sym + car
    return sym

def sym_plateau(plateau):
    """renvoie le symétrique du plateau de jeu par rapport à l'axe vertical"""
    p = copy.deepcopy(plateau)
    for i in range(3):
        temp = p[i][0]
        p[i][0] = p[i][2]
        p[i][2] = temp
    return p

def donne_couleur(rang):
    couleur = ''
    if rang % 2 == 0:
        couleur = 'B'
    elif rang % 2 == 1:
        couleur = 'N'
    return couleur

def etat_plateau(liste_coups):
    """renvoie l'état d'un plateau après la liste des coups joués"""
    plateau = [["N","N","N"],[" "," "," "],["B","B","B"]]
    for i in range(len(liste_coups)):
        coup = chaine_vers_coord(liste_coups[i])
        if i%2 == 0:
            couleur = "B"
        else:
            couleur = "N"
        plateau[coup[0][0]][coup[0][1]] = " "
        plateau[coup[1][0]][coup[1][1]] = couleur
    return plateau



def positions(plateau, couleur):
    """dresse la liste des positions des pions de la couleur donnée dans un plateau donné"""
    liste_positions = []
    for i in range(3):
        for j in range(3):
            if plateau[i][j] == couleur:
                liste_positions = liste_positions + [[i,j]]
    return liste_positions




# exemple = [["N"," ","N"],[" ","N","B"],["B","B"," "]];
# affichage(exemple);
# positions(exemple,"N") # les valeurs renvoyés sont les coordonnées des pions noirs




def coups_possibles(plateau, couleur_a_jouer):
    """dresse la liste des mouvements possibles sur le plateau pour la couleur qui doit jouer"""
    liste_coups = []
    adversaire = autre_couleur(couleur_a_jouer)
    dir = direction(couleur_a_jouer)
    for pos in positions(plateau, couleur_a_jouer):
        i = pos[0]
        j = pos[1]
        if i != rang_initial(adversaire): # cela veut dire qu'on peut encore avancer
            if plateau[i+dir][j] == " ":
                liste_coups = liste_coups + [[pos,[i+dir,j]]]
            if j < 2 and plateau[i+dir][j+1] == adversaire:
                liste_coups = liste_coups + [[pos,[i+dir,j+1]]]
            if j > 0 and plateau[i+dir][j-1] == adversaire:
                 liste_coups = liste_coups + [[pos,[i+dir,j-1]]]
    return sorted(liste_coups, key = lambda couple : (couple[0][1],couple[1][1]))




# exemple = [["N"," ","N"],[" ","N","B"],["B","B"," "]];
# affichage(exemple);
# coups_possibles(exemple,"N")





def plateau_joue(plateau, couleur_a_jouer, coup):
    """renvoie le plateau après le coup joué par la couleur"""
    p = copy.deepcopy(plateau)
    if coup in coups_possibles(plateau, couleur_a_jouer):
        p[coup[0][0]][coup[0][1]] = " "
        p[coup[1][0]][coup[1][1]] = couleur_a_jouer
    return p





# exemple = [["N"," ","N"],[" ","N","B"],["B","B"," "]];
# affichage(exemple);
# mouvement = [[1, 1], [2, 0]];
# exemple_apres_coup = plateau_joue(exemple, "N", mouvement);
# affichage(exemple_apres_coup)





def fin_de_partie(plateau, couleur_jouee):
    """teste la victoire des pions de la couleur indiquée, en supposant que ceux-ci viennent de jouer"""
    partie_finie = False
    adversaire = autre_couleur(couleur_jouee)
    if positions(plateau, adversaire) == []:
        partie_finie = True
    for pos in positions(plateau, couleur_jouee):
        i = pos[0]
        j = pos[1]
        if i == rang_initial(adversaire):
            partie_finie = True
    if  len(positions(plateau, adversaire)) > 0 and len(coups_possibles(plateau, adversaire)) == 0:
        partie_finie = True
    return partie_finie




# fin_de_partie(exemple_apres_coup, "N") # Victoire des noirs ?


# # Jeu entre deux joueurs aléatoires



import random
def partie_alea_vs_alea():
    """renvoie une partie aléatoire entre deux joueurs qui jouent au hasard"""
    tour = 0
    plateau = [["N","N","N"],[" "," "," "],["B","B","B"]]
    vainqueur = ""
    partie_en_cours = True
    while partie_en_cours == True:
        if tour % 2 == 0: #au tour des blancs de jouer
            coup_choisi = random.choice(coups_possibles(plateau,"B")) # choix d'un coup au hasard
            plateau = plateau_joue(plateau, 'B', coup_choisi) # coup joué
            if fin_de_partie(plateau, "B") == True: # vérification d'une éventuelle victoire des blancs
                partie_en_cours = False # arrêt de la partie
                vainqueur = "B"
        if tour % 2 == 1: #au tour des noirs de jouer
            coup_choisi = random.choice(coups_possibles(plateau,"N")) # choix d'un coup au hasard
            plateau = plateau_joue(plateau, 'N', coup_choisi) # coup joué
            if fin_de_partie(plateau, "N") == True: # vérification d'une éventuelle victoire des noirs
                partie_en_cours = False # arrêt de la partie
                vainqueur = "N"
        tour = tour + 1 # on passe au tour suivant
    return vainqueur




# partie_alea_vs_alea() # appel de la fonction pour réaliser une partie aléatoire




def echantillon(nb_parties):
    """renvoie la fréquence de victoire de chaque couleur après simulation de n parties"""
    victoires_blancs = 0
    victoires_noirs = 0
    for _ in range(nb_parties):
        partie = partie_alea_vs_alea()
        if partie == 'B':
            victoires_blancs = victoires_blancs + 1
        if partie == 'N':
            victoires_noirs = victoires_noirs + 1
    return victoires_blancs / nb_parties, victoires_noirs / nb_parties





# echantillon(5000) #




import matplotlib.pyplot as plt
def graphique(liste_vainqueurs):
    """On représente graphiquement une liste de vainqueurs"""
    plt.close ('all')
    abscisses = [0]
    ordonnees = [0]
    for k in range(1,len(liste_vainqueurs)):
        abscisses = abscisses + [abscisses[k-1]+1]
        if liste_vainqueurs[k] == "N":
            ordonnees = ordonnees + [ordonnees[k-1]+1]
        if liste_vainqueurs[k] == "B":
            ordonnees = ordonnees + [ordonnees[k-1]-1]
    fig = plt.figure()
    plt.plot(abscisses, ordonnees)
    plt.show()


# In[20]:


# graphique([partie_alea_vs_alea() for _ in range(500)])




def arbre_de_jeu():
    """construit l'arbre de jeu complet d'hexapawn"""
    parties_completes = []
    plateau_initial = [["N","N","N"],[" "," "," "],["B","B","B"]]
    parties_en_cours = [[]]
    while parties_en_cours != []:
        partie = parties_en_cours[0]
        longueur = len(partie)
        couleur_a_jouer = donne_couleur(longueur)
        plateau = etat_plateau(partie)
        possibilites = coups_possibles(plateau, couleur_a_jouer)
        for coup in possibilites:
            suite_partie = partie + [coord_vers_chaine(coup)]
            plateau_suivant = etat_plateau(suite_partie)
            if fin_de_partie(plateau_suivant, couleur_a_jouer) == True:
                parties_completes = parties_completes + [suite_partie]
            else:
                parties_en_cours = parties_en_cours + [suite_partie]
        parties_en_cours.pop(0)
    return sorted(parties_completes)





# print(arbre_de_jeu()[37]); # affiche un exemple de partie pris dans l'arbre de jeu
# len(arbre_de_jeu()) # affiche le nombres de parties possibles


# ## Création de la pioche et d'une fonction de recherche dans celle-ci



def boites():
    """renvoie la matrice des états de jeux distincts et des coups possibles pour l'ensemble des branches de l'arbre et toutes les étapes"""
    arbre = arbre_de_jeu()
    etats = [[[etat_plateau(partie[:j]),coups_possibles(etat_plateau(partie[:j]),donne_couleur(j)) ] for j in range(len(partie))] for partie in arbre ]
    extraction = [[] for i in range(7)]
    for partie in etats:
        for j in range(len(partie)):
            situation = partie[j]
            if situation not in extraction[j]:
                extraction[j] = extraction[j] + [situation]
    longueurs = [len(extraction[j]) for j in range(7)] # pour vérifier le nombre de boites différentes
    return extraction

# variable globale contenant toutes les boites
PIOCHE = boites()

# pour activer la recherche
def recherche(plateau,tour):
    """recherche dans la liste des états possibles le plateau obtenu au tour donné"""
    global PIOCHE
    for i in range(len(PIOCHE[tour])):
        if PIOCHE[tour][i][0] == plateau:
            return i




# boites()[3][4][0] # pour un tour des noirs (rang 3, impaire), situation de jeu numérotée 4 dans la pioche



# recherche([['N', 'N', ' '], ['B', 'B', 'N'], [' ', ' ', 'B']],3) # retrouve le numéro de la situation qui correspond

def sequence_alea_vs_ia(nb_parties, punition = True, recompense = False):
    """renvoie une serie de nb_parties parties d'hexapawn entre un joueur aléatoire et une IA qui valorise ou élimine ses coups selon l'issue de la partie et le type de feedback choisi"""
    pioche = boites() # etats de jeux avec plateaux et coups possibles
    liste_vainqueurs = []
    liste_parcours = []
    for _ in range(nb_parties): # on simule n parties et on élimine les parties perdantes pour l'IA
        tour = 0
        plateau = [["N","N","N"],[" "," "," "],["B","B","B"]]
        partie_en_cours = True
        dernier_coup_noirs = []
        dernier_coup_noirs_sym = []
        while partie_en_cours == True:
            if tour % 2 == 0: # au tour des blancs de jouer
                coup = random.choice(coups_possibles(plateau,"B")) # choix d'un coup au hasard
                plateau = plateau_joue(plateau, 'B', coup) # coup joué
                if fin_de_partie(plateau, "B") == True: # vérification d'une éventuelle victoire des blancs
                    partie_en_cours = False # arrêt de la partie
                    liste_vainqueurs = liste_vainqueurs + ["B"]
                    if punition:
                        rang, partie, mouvement = dernier_coup_noirs.pop() # rappel du dernier coup noir
                        rang_sym, partie_sym, mouvement_sym = dernier_coup_noirs_sym.pop()  # et de son symétrique
                        pioche[rang][partie][1].remove(mouvement) # suppression de ce choix perdant pour l'IA (punition)
                        pioche[rang_sym][partie_sym][1].remove(mouvement_sym)
            if tour % 2 == 1: # au tour des noirs de jouer IA
                partie = recherche(plateau,tour) # recherche de la partie dans les états de jeux
                partie_symetrique = recherche(sym_plateau(plateau),tour) # recherche de la partie symétrique
                if pioche[tour][partie][1] == []: # "boite" vide donc les noirs ne peuvent pas jouer et on recherche le dernier mouvement noir à éliminer
                    partie_en_cours = False # arrêt de la partie
                    liste_vainqueurs = liste_vainqueurs + ["B"] # ajout du vainqueur blanc à la liste des vainqueurs
                    if punition:
                        rang, partie, mouvement = dernier_coup_noirs.pop()
                        rang_sym, partie_sym, mouvement_sym = dernier_coup_noirs_sym.pop()
                        pioche[rang][partie][1].remove(mouvement) # suppression du dernier mouvement
                        pioche[rang_sym][partie_sym][1].remove(mouvement_sym)
                else:
                    coup = random.choice(pioche[tour][partie][1]) # l'IA choisit un coup au hasard dans la boite correpondant à l'état de jeu
                    dernier_coup_noirs.append([tour, partie, coup])  # on enregistre le dernier coup noir (fonctionnement comme une pile, ici on empile
                    dernier_coup_noirs_sym.append([tour, partie_symetrique, sym_mouv(coup)])
                    plateau = plateau_joue(plateau, 'N', coup)
                    if fin_de_partie(plateau, "N") == True:
                        partie_en_cours = False
                        liste_vainqueurs = liste_vainqueurs + ["N"]
                        if recompense: # on recompense le dernier mouvement gagnant en le rajoutant plusieurs fois dans la pioche afin d'augmenter sa probabilité d'être choisi aux prochains tours
                            rajout = 3 # le nombre de coups gagnants rajouté peut être modifié ici
                            for _ in range(rajout):
                                pioche[tour][partie][1].append(coup)
                                pioche[tour][partie_symetrique][1].append(sym_mouv(coup))
            tour = tour + 1
    return liste_vainqueurs



# print(sequence_alea_vs_ia(100))




# graphique(sequence_alea_vs_ia(100, punition = False, recompense = False))
#
#
#
#
# graphique(sequence_alea_vs_ia(100, punition = False, recompense = True))
#
#
#
#
# graphique(sequence_alea_vs_ia(500, punition = False, recompense = True))
#
#
#
#
# graphique(sequence_alea_vs_ia(50, punition = True, recompense = False))
#
#

#
#
# graphique(sequence_alea_vs_ia(200, punition = True, recompense = False))





# graphique(sequence_alea_vs_ia(50, punition = True, recompense = True))





# graphique(sequence_alea_vs_ia(200, punition = True, recompense = True))




def recherche_strategie(nb_parties, punition = True, recompense = False):
    """renvoie une serie de nb_parties parties d'hexapawn entre un joueur aléatoire et une IA qui valorise ou élimine ses coups selon l'issue de la partie et le type de feedback choisi"""
    pioche = boites() # etats de jeux avec plateaux et coups possibles
    liste_vainqueurs = []
    liste_parcours = []
    liste_parties_gagnees_noirs = []
    for _ in range(nb_parties): # on simule n parties et on élimine les parties perdantes pour l'IA
        tour = 0
        plateau = [["N","N","N"],[" "," "," "],["B","B","B"]]
        partie_en_cours = True
        dernier_coup_noirs = []
        dernier_coup_noirs_sym = []
        liste_mouvements = []
        while partie_en_cours == True:
            if tour % 2 == 0: # au tour des blancs de jouer
                coup = random.choice(coups_possibles(plateau,"B")) # choix d'un coup au hasard
                liste_mouvements = liste_mouvements + [coord_vers_chaine(coup)]
                plateau = plateau_joue(plateau, 'B', coup) # coup joué
                if fin_de_partie(plateau, "B") == True: # vérification d'une éventuelle victoire des blancs
                    partie_en_cours = False # arrêt de la partie
                    liste_vainqueurs = liste_vainqueurs + ["B"]
                    if punition:
                        rang, partie, mouvement = dernier_coup_noirs.pop() # rappel du dernier coup noir
                        rang_sym, partie_sym, mouvement_sym = dernier_coup_noirs_sym.pop()  # et de son symétrique
                        pioche[rang][partie][1].remove(mouvement) # suppression de ce choix perdant pour l'IA (punition)
                        pioche[rang_sym][partie_sym][1].remove(mouvement_sym)
            if tour % 2 == 1: # au tour des noirs de jouer IA
                partie = recherche(plateau,tour) # recherche de la partie dans les états de jeux
                partie_symetrique = recherche(sym_plateau(plateau),tour) # recherche de la partie symétrique
                if pioche[tour][partie][1] == []: # "boite" vide donc les noirs ne peuvent pas jouer et on recherche le dernier mouvement noir à éliminer
                    partie_en_cours = False # arrêt de la partie
                    liste_vainqueurs = liste_vainqueurs + ["B"] # ajout du vainqueur blanc à la liste des vainqueurs
                    if punition:
                        rang, partie, mouvement = dernier_coup_noirs.pop()
                        rang_sym, partie_sym, mouvement_sym = dernier_coup_noirs_sym.pop()
                        pioche[rang][partie][1].remove(mouvement) # suppression du dernier mouvement
                        pioche[rang_sym][partie_sym][1].remove(mouvement_sym)
                else:
                    coup = random.choice(pioche[tour][partie][1]) # l'IA choisit un coup au hasard dans la boite correpondant à l'état de jeu
                    dernier_coup_noirs.append([tour, partie, coup])  # on enregistre le dernier coup noir (fonctionnement comme une pile, ici on empile
                    dernier_coup_noirs_sym.append([tour, partie_symetrique, sym_mouv(coup)])
                    plateau = plateau_joue(plateau, 'N', coup)
                    liste_mouvements = liste_mouvements + [coord_vers_chaine(coup)]
                    if fin_de_partie(plateau, "N") == True:
                        partie_en_cours = False
                        liste_vainqueurs = liste_vainqueurs + ["N"]
                        if liste_mouvements not in liste_parties_gagnees_noirs: # ajout de la partie gagnante
                            liste_parties_gagnees_noirs = liste_parties_gagnees_noirs + [liste_mouvements]
                        if recompense: # on recompense le dernier mouvement gagnant en le rajoutant plusieurs fois dans la pioche afin d'augmenter sa probabilité d'être choisi aux prochains tours
                            rajout = 3 # le nombre de coups gagnants rajouté peut être modifié ici
                            for _ in range(rajout):
                                pioche[tour][partie][1].append(coup)
                                pioche[tour][partie_symetrique][1].append(sym_mouv(coup))

            tour = tour + 1
    return sorted(liste_parties_gagnees_noirs), len(liste_parties_gagnees_noirs)




# recherche_strategie(10000)




# recherche_strategie(10000, recompense = True)





